{ Original code from D1P }
unit tcIconSize;

interface

uses
  Windows, Classes, SysUtils, IniFiles, Multimon, tcSpecificFunctions;

function GetTCIconsSize(DefaultSize: Integer): Integer;
function GetTCIconsSizeEx(DefaultSize: Integer): Integer;

type
  MONITOR_DPI_TYPE = (
    MDT_EFFECTIVE_DPI = 0,
    MDT_ANGULAR_DPI = 1,
    MDT_RAW_DPI = 2,
    MDT_DEFAULT = MDT_EFFECTIVE_DPI);
  TMonitorDpiType = MONITOR_DPI_TYPE;

function GetDpiForMonitor(hmonitor: HMONITOR; dpiType: TMonitorDpiType; out dpiX: UINT; out dpiY: UINT): HRESULT; stdcall;

implementation

function GetDpiForMonitor; external 'Shcore.dll' name 'GetDpiForMonitor' delayed;


function GetScreenDPI(MonitorHandle: HMONITOR = 0): Integer;
var
  DC: HDC;
  Ydpi, Xdpi: Cardinal;
begin
  if (MonitorHandle > 0) and CheckWin32Version(6,3) then
  begin
    if GetDpiForMonitor(MonitorHandle, TMonitorDpiType.MDT_EFFECTIVE_DPI, Ydpi, Xdpi) = S_OK then
      Result := Ydpi
    else
      Result := 0;
  end else begin
    DC := GetDC(0);
    Result := GetDeviceCaps(DC, LOGPIXELSY);
    ReleaseDC(0, DC);
  end;
end;

function FindTCWindow: HWND;
begin
  Result := FindWindow('TTOTAL_CMD', nil);
end;

function FindTCIniFileName: string;
begin
  Result := GetEnvironmentVariable('COMMANDER_INI');
end;

function GetTCIconsSize(DefaultSize: Integer): Integer;
var
  TC_INI: TIniFile;
  ResolutionSpecific: Boolean;
  IconsSizeSectionName, IniFileName: string;
  MonInfo: TMonitorInfo;
begin
  Result := DefaultSize;
  IniFileName := FindTCIniFileName;
  if FileExists(IniFileName) then
  begin
    TC_INI := TIniFile.Create(IniFileName);
    try
      ResolutionSpecific := TC_INI.ReadBool('Configuration',
        'ResolutionSpecific', True);
      if ResolutionSpecific then
      begin
        MonInfo.cbSize := SizeOf(MonInfo);
        GetMonitorInfo(MonitorFromWindow(FindTCWindow,
          MONITOR_DEFAULTTONEAREST), @MonInfo);
        IconsSizeSectionName :=
          Format('%dx%d', [MonInfo.rcMonitor.Right - MonInfo.rcMonitor.Left,
          MonInfo.rcMonitor.Bottom - MonInfo.rcMonitor.Top]) + ' (8x16)';
        // normal font section
        if not TC_INI.SectionExists(IconsSizeSectionName) then
        begin
          IconsSizeSectionName :=
            Format('%dx%d', [MonInfo.rcMonitor.Right - MonInfo.rcMonitor.Left,
            MonInfo.rcMonitor.Bottom - MonInfo.rcMonitor.Top]) + ' (10x20)';
          // large font section
          if not TC_INI.SectionExists(IconsSizeSectionName) then
            IconsSizeSectionName := 'AllResolutions';
        end;
      end
      else
        IconsSizeSectionName := 'AllResolutions';

      Result := TC_INI.ReadInteger(IconsSizeSectionName, 'Iconsize32', Result);
    finally
      TC_INI.Free;
    end;
  end;
end;

function GetTCIconsSizeEx(DefaultSize: Integer): Integer;

  function SectionExists(const Section: string): Boolean;
  var
    INI: TIniFile;
  begin
    Result := False;
    INI := GetTCINIWithRedirectedSection(Section);
    try
      if Assigned(INI) then
        Result := INI.SectionExists(Section);
    finally
      FreeAndNil(INI);
    end;
  end;

const
  tcScreenSections: array [Boolean] of string = (' (8x16)', ' (10x20)');
var
  TC_INI: TIniFile;
  ResolutionSpecific: Boolean;
  IconsSizeSectionName: string;
  MonInfo: TMonitorInfo;
  ScreenDPI: Integer;
  MonitorHandle: HMONITOR;
begin
  Result := DefaultSize;

  TC_INI := GetTCINIWithRedirectedSection('Configuration');

  if Assigned(TC_INI) then
  try
    ResolutionSpecific := TC_INI.ReadBool('Configuration',
      'ResolutionSpecific', True);
    if ResolutionSpecific then
    begin
      MonInfo.cbSize := SizeOf(MonInfo);

      MonitorHandle := MonitorFromWindow(FindTCWindow, MONITOR_DEFAULTTONEAREST);

      if GetMonitorInfo(MonitorHandle, @MonInfo) then
      begin
        ScreenDPI := GetScreenDPI(MonitorHandle);

        IconsSizeSectionName :=
          Format('%dx%d', [MonInfo.rcMonitor.Right - MonInfo.rcMonitor.Left,
          MonInfo.rcMonitor.Bottom - MonInfo.rcMonitor.Top]) + tcScreenSections[ScreenDPI > 96];
        if not SectionExists(IconsSizeSectionName) then
        begin
          IconsSizeSectionName :=
            Format('%dx%d', [MonInfo.rcMonitor.Right - MonInfo.rcMonitor.Left,
            MonInfo.rcMonitor.Bottom - MonInfo.rcMonitor.Top]) + tcScreenSections[not (ScreenDPI > 96)];
          if not SectionExists(IconsSizeSectionName) then
            IconsSizeSectionName := 'AllResolutions';
        end;

        {
        IconsSizeSectionName :=
          Format('%dx%d', [MonInfo.rcMonitor.Right - MonInfo.rcMonitor.Left,
          MonInfo.rcMonitor.Bottom - MonInfo.rcMonitor.Top]) + ' (8x16)';
        // normal font section
        if not SectionExists(IconsSizeSectionName) then
        begin
          IconsSizeSectionName :=
            Format('%dx%d', [MonInfo.rcMonitor.Right - MonInfo.rcMonitor.Left,
            MonInfo.rcMonitor.Bottom - MonInfo.rcMonitor.Top]) + ' (10x20)';
          // large font section
          if not SectionExists(IconsSizeSectionName) then
            IconsSizeSectionName := 'AllResolutions';
        end;
        }
      end else
        IconsSizeSectionName := 'AllResolutions';

    end else
      IconsSizeSectionName := 'AllResolutions';

    FreeAndNil(TC_INI);
    TC_INI := GetTCINIWithRedirectedSection(IconsSizeSectionName);
    if Assigned(TC_INI) then
      Result := TC_INI.ReadInteger(IconsSizeSectionName, 'Iconsize32', Result);

  finally
    TC_INI.Free;
  end;
end;

end.
